function Shape3D(fov)
	local shape = {
		vertices = {},
		_transformed = {},
		rotation = {},
		camera = {0, 0, 80},
		averageZ = 0,
		colors = {},
		shader = nil,
		position = {0, 0, 0},
		lights = {},
		light_c = {},
		fov = fov
	}

	function shape:RecalculateVertices()
		local w, h = love.graphics.getWidth(), love.graphics.getHeight()

		for i, v in ipairs(self.vertices) do
			local factor = math.deg(self.fov) / (self.camera[3] + v[3] + self.position[3])

			local x = (v[1] * factor) + (w / 2) + self.position[1]
			local y = (-v[2] * factor) + (h / 2) + self.position[2]

			self.averageZ = self.averageZ + v[3]

			self._transformed[i] = {x, y, v[3]}
			--[[if i % 3 == 0 then
				table.insert(self.colors, {0, 0, 0})
			end]]
		end

		if self.shader ~= nil and #self.lights > 0 then
			self.shader:send("lights", unpack(self.lights))
			self.shader:send("light_colors", unpack(self.light_c))
			self.shader:send("light_count", #self.lights)
		end

		self.averageZ = self.averageZ / #self.vertices
	end

	function shape:ApplyShader(shaderSource)
		self.shader = love.graphics.newPixelEffect(shaderSource)
	end

	function shape:AddLight(x, y, r, g, b)
		if not r then r = 255 end
		if not g then g = 255 end
		if not b then b = 255 end

		table.insert(self.lights, {x, y})
		table.insert(self.light_c, {r / 255, g / 255, b / 255})
	end

	function shape:DrawZOrdered()
		if #self._transformed == 0 then return end

		local faces = {minY = 0}

		function table.contains(t, c)
			for i, v in ipairs(t) do
				if v[1] == c[1] and v[2] == c[2] and v[3] == c[3] then return true end
			end

			return false
		end

		local id = 0

		faces.minY = math.huge

		for i = 1, #self._transformed, 6 do
			local face = {}

			local vertex1 = self._transformed[i + 0]
			local vertex2 = self._transformed[i + 1]
			local vertex3 = self._transformed[i + 2]
			local vertex4 = self._transformed[i + 3]
			local vertex5 = self._transformed[i + 4]
			local vertex6 = self._transformed[i + 5]

			local verts = {vertex1, vertex2, vertex3, vertex4, vertex5, vertex6}

			for i, v in ipairs(verts) do
				if not table.contains(face, v) then
					table.insert(face, {v[1], v[2], v[3]})
				end
			end

			id = id + 1

			local avgY = (face[1][2] + face[2][2] + face[3][2] + face[4][2]) / 4
			faces.minY = math.min(faces.minY, avgY)

			table.insert(faces, {
				face = face, 
				id = id,
				averageZ = (face[1][3] + face[2][3] + face[3][3] + face[4][3]) / 4,
				averageY = avgY,
				averageX = (face[1][1] + face[2][1] + face[3][1] + face[4][1]) / 4})
		end

		table.sort(faces, function(a, b)
			return a.averageZ > b.averageZ
		end)

		if self.shader ~= nil then
			love.graphics.setPixelEffect(self.shader)

			-- self.shader:send("x", love.mouse.getX())
			-- self.shader:send("y", love.mouse.getY())
		end

		for i,v in ipairs(faces) do
			local vert1 = v.face[1]
			local vert2 = v.face[2]
			local vert3 = v.face[3]
			local vert4 = v.face[4]

			if self.shader ~= nil then
				self.shader:send("fx", v.averageX)
				self.shader:send("fy", v.averageY)
			end

			love.graphics.polygon("fill", 
				vert1[1], vert1[2],
				vert2[1], vert2[2],
				vert3[1], vert3[2],
				vert4[1], vert4[2])
		end

		love.graphics.setPixelEffect(nil)
	end

	function shape:AddVertex(x, y, z)
		table.insert(self.vertices, {x, y, z})
	end

	function shape:Draw()
		local index = 0
		for i = 1, #self._transformed, 3 do
			index = index + 1
			love.graphics.setColor(
				self.colors[index][1] + self.averageZ * 5, 
				self.colors[index][2] + self.averageZ * 5, 
				self.colors[index][3] + self.averageZ * 5)

			local x1, x2, x3 = self._transformed[i][1], self._transformed[i + 1][1], self._transformed[i + 2][1]
			local y1, y2, y3 = self._transformed[i][2], self._transformed[i + 1][2], self._transformed[i + 2][2]

			love.graphics.polygon("fill", x1, y1, x2, y2, x3, y3)
		end
	end

	function shape:RotateX(deg)
		local rads = math.rad(deg)
		local cosa = math.cos(rads)
		local sina = math.sin(rads)

		for i,v in ipairs(self.vertices) do
			local y = v[2] * cosa - v[3] * sina
			local z = v[2] * sina + v[3] * cosa

			self.vertices[i] = {v[1], y, z}
		end
	end

	function shape:RotateY(deg)
		local rads = math.rad(deg)
		local cosa = math.cos(rads)
		local sina = math.sin(rads)

		for i,v in ipairs(self.vertices) do
			local z = v[3] * cosa - v[1] * sina
			local x = v[3] * sina + v[1] * cosa

			self.vertices[i] = {x, v[2], z}
		end
	end

	function shape:RotateZ(deg)
		local rads = math.rad(deg)
		local cosa = math.cos(rads)
		local sina = math.sin(rads)

		for i,v in ipairs(self.vertices) do
			local x = v[1] * cosa - v[2] * sina
			local y = v[1] * sina + v[2] * cosa

			self.vertices[i] = {x, y, v[3]}
		end
	end

	return shape
end

function Cube(x, y, z)
	local cube = Shape3D(45)
	local x, y, z = x / 2, y / 2, z / 2

	-- bottom face
	cube:AddVertex(-x, -y, -z)
	cube:AddVertex(-x, -y, z)
	cube:AddVertex(x, -y, z)

	cube:AddVertex(-x, -y, -z)
	cube:AddVertex(x, -y, -z)
	cube:AddVertex(x, -y, z)

	-- top face
	cube:AddVertex(-x, y, -z)
	cube:AddVertex(-x, y, z)
	cube:AddVertex(x, y, z)

	cube:AddVertex(-x, y, -z)
	cube:AddVertex(x, y, -z)
	cube:AddVertex(x, y, z)

	-- left face
	cube:AddVertex(-x, -y, -z)
	cube:AddVertex(-x, -y, z)
	cube:AddVertex(-x, y, z)

	cube:AddVertex(-x, -y, -z)
	cube:AddVertex(-x, y, -z)
	cube:AddVertex(-x, y, z)

	-- right face
	cube:AddVertex(x, -y, -z)
	cube:AddVertex(x, -y, z)
	cube:AddVertex(x, y, z)

	cube:AddVertex(x, -y, -z)
	cube:AddVertex(x, y, -z)
	cube:AddVertex(x, y, z)

	-- front face
	cube:AddVertex(-x, -y, -z)
	cube:AddVertex(-x, y, -z)
	cube:AddVertex(x, y, -z)

	cube:AddVertex(-x, -y, -z)
	cube:AddVertex(x, -y, -z)
	cube:AddVertex(x, y, -z)

	-- back face
	cube:AddVertex(-x, -y, z)
	cube:AddVertex(-x, y, z)
	cube:AddVertex(x, y, z)
	
	cube:AddVertex(-x, -y, z)
	cube:AddVertex(x, -y, z)
	cube:AddVertex(x, y, z)

	cube:RecalculateVertices()

	return cube
end

local c

function love.load()
	c = Cube(10, 10, 10)

	c:ApplyShader(love.filesystem.read("shader.fs"))
	c:AddLight(love.graphics.getWidth() / 2, 0, 255, 255, 255)
	c:AddLight(love.graphics.getWidth() / 3, 300, 0, 128, 255)
end

function love.update(dt)
	c:RotateY(dt * 15)
	c:RotateX(dt * 5)
	c:RecalculateVertices()

	love.graphics.setCaption(love.timer.getFPS())
end

function love.draw()
	c:DrawZOrdered()
end
